import math
import random
from collections.abc import Callable

from pydantic import BaseModel, Field

from massranker.elo import expected, elo

class ItemT(BaseModel):
    name: str
    elo: float
    enabled: bool
    times_seen: int = Field(default=0)


class StateT(BaseModel):
    items: list[ItemT]

    @classmethod
    def default(cls):
        return StateT(items=[])

    def _bring_up_to_date(self, names: list[str]):
        names_set = set(names)
        seen = set()

        # == first: figure out the list of enabled names ==
        for i in self.items:
            i.enabled = i.name in names_set
            seen.add(i.name)

        remaining_names = names_set - seen
        for name in remaining_names:
            self.items.append(ItemT(
                name=name,
                elo=1500,
                enabled=True,
            ))

    def do_tournament(self, names: list[str], callback: Callable[[list[ItemT]], list[ItemT]]):
        # bring up to date
        self._bring_up_to_date(names)

        # set up a tournament
        possible_entrants = [i for i in self.items if i.enabled]
        if len(possible_entrants) < 2:
            raise ValueError("need at least two things to compare")

        # build a sample
        sample_size = min(len(possible_entrants), 8)
        if sample_size == len(possible_entrants):
            entrants = random.sample(possible_entrants, k=sample_size)
        else:
            # this is a messy way to take a weighted random sample
            picked = []
            while len(picked) < sample_size:
                remaining_possible_entrants = [i for i in possible_entrants if i not in picked]
                weights = [1/(e.times_seen + 1) for e in remaining_possible_entrants]
                for i in random.choices(remaining_possible_entrants, weights=weights, k=sample_size - len(picked)):
                    if i not in picked:
                        picked.append(i)
            random.shuffle(picked)
            entrants = picked

        expected_score = [0.0] * len(entrants)
        score = [0.0] * len(entrants)

        reranked = callback(list(entrants))

        for e, entrant in enumerate(entrants):
            for e2, other in enumerate(entrants):
                if e2 == e:
                    continue
                expected_score[e] += expected(entrant.elo, other.elo)
                score[e] += float(reranked.index(entrant) < reranked.index(other))

        for e, entrant in enumerate(entrants):
            entrant.elo = elo(entrant.elo, expected_score[e], score[e])
            entrant.times_seen += 1

    def dump_rankings(self) -> list[tuple[str, int]]:
        return [
            (e.name, int(e.elo))
            for e in sorted(self.items, key=lambda e: e.elo, reverse=True)
            if e.enabled
        ]